"""
Merkle-tree core functionality
"""

from abc import ABCMeta, abstractmethod
from collections import deque, namedtuple
from threading import Lock
import builtins

from cachetools import LRUCache

from pymerkle.hasher import MerkleHasher
from pymerkle.proof import MerkleProof
from pymerkle.utils import log2, decompose


try:
    builtins.profile
except AttributeError:
    def profile(func):
        return func

    builtins.profile = profile


class InvalidChallenge(Exception):
    """
    Raised when no Merkle-proof exists for the provided prameters
    """
    pass



_CacheInfo = namedtuple('CacheInfo', ['size', 'capacity', 'hits', 'misses'])


class BaseMerkleTree(MerkleHasher, metaclass=ABCMeta):
    """
    Abstract base class encapsulating the core Merkle-tree functionalities in
    a storage agnostic fashion.

    .. note:: Concrete implentations should inherit from this class and
        implement its interior storage interface.

    :param algorithm: [optional] hash algorithm. Defaults to *sha256*.
    :type algorithm: str
    :param threshold: [optional] Subroot cache threshold. Defaults to 128.
    :type threshold: int
    :param capacity: [optional] Subroot cache capacity in bytes. Defaults to
        1GiB.
    :type capacity: int
    :param disable_security: [optional] if *True*, resistance against
        second-preimage attack will be deactivated. Defaults to *False*.
    :type disable_security: boolean
    :param disable_optimizations: [optional] if *True*, low-level operations
        will fallback to their unoptimized recursive version. Defaults to
        *False*.
    :type disable_optimizations: boolean
    :param disable_optimizations: [optional] if *True*, low-level operations
        will fallback to their unoptimized recursive version. Defaults to
        *False*.
    :type disable_optimizations: boolean
    :param disable_cache: [optional] if *True*, subroot caching will be
        deactivated. Defaults to *False*.
    :type cache: boolean
    """

    def __init__(self, algorithm='sha256', **opts):
        self.algorithm = algorithm
        self.security = not opts.get('disable_security', False)
        self.threshold = opts.get('threshold', 128)
        self.capacity = opts.get('capacity', 1024 ** 3)
        self.cache = LRUCache(maxsize=self.capacity, getsizeof=len)
        self.hits = 0
        self.misses = 0
        self.lock = Lock()

        if opts.get('disable_optimizations', False):
            self._get_root = self._get_root_naive
            self._inclusion_path = self._inclusion_path_naive
            self._consistency_path = self._consistency_path_naive

        if opts.get('disable_cache', False):
            self._get_subroot = self._get_subroot_uncached

        super().__init__(self.algorithm, self.security)


    def _hash_entry(self, data):
        return self.hash_buff(data)


    def _hash_nodes(self, lnode, rnode):
        return self.hash_pair(lnode, rnode)


    def append_entry(self, data):
        """
        Appends a new leaf storing the provided data entry.

        :param data: data to append
        :type data: whatever expected according to application logic
        :returns: index of newly appended leaf counting from one
        :rtype: int
        """
        buffer = self._encode_entry(data)
        digest = self._hash_entry(buffer)
        index = self._store_leaf(data, digest)

        return index


    def get_leaf(self, index):
        """
        Returns the leaf hash located at the provided position.

        :param index: leaf index counting from one
        :type index: int
        :rtype: bytes
        """
        return self._get_leaf(index)


    def get_size(self):
        """
        Returns the current number of leaves.

        :rtype: int
        """
        return self._get_size()


    def get_state(self, size=None):
        """
        Computes the root-hash of the tree corresponding to the provided number
        of leaves.

        :param size: [optional] number of leaves to consider. Defaults to
            current tree size.
        :type size: int
        :rtype: bytes
        """
        if size is None:
            size = self._get_size()

        return self._get_root(0, size)


    def prove_inclusion(self, index, size=None):
        """
        Proves inclusion of the hash located at the provided index against the
        tree corresponding to the provided number of leaves.

        :param index: leaf index counting from one
        :type index: int
        :param size: [optional] number of leaves to consider. Defaults to
            current tree size
        :type size: int
        :rtype: MerkleProof
        :raises InvalidChallenge: if the provided parameters are invalid or
            incompatible with each other
        """
        currsize = self.get_size()

        if size is None:
            size = currsize

        if not (0 < size <= currsize):
            raise InvalidChallenge('Provided size is out of bounds')

        if not (0 < index <= size):
            raise InvalidChallenge('Provided index is out of bounds')

        rule, path = self._inclusion_path(0, index - 1, size, 0)

        return MerkleProof(self.algorithm, self.security, size, rule, [],
                path)


    def prove_consistency(self, size1, size2=None):
        """
        Proves consistency between the states corresponding to the provided
        sizes.

        :param size1: number of leaves for prior state
        :type size1: int
        :param size2: [optional] number of leaves for later state. Defaults to
            current tree size.
        :type size2: int
        :rtype: MerkleProof
        :raises InvalidChallenge: if the provided parameters are invalid or
            incompatible with each other
        """
        currsize = self.get_size()

        if size2 is None:
            size2 = currsize

        if not (0 < size2 <= currsize):
            raise InvalidChallenge('Provided later size out of bounds')

        if not (0 < size1 <= size2):
            raise InvalidChallenge('Provided prior size out of bounds')

        rule, subset, path = self._consistency_path(0, size1, size2, 0)

        return MerkleProof(self.algorithm, self.security, size2, rule,
                subset, path)


    def get_cache_info(self):
        """
        Returns subroot cache info.
        """
        return _CacheInfo(self.cache.currsize, self.cache.maxsize, self.hits,
                self.misses)


    def cache_clear(self):
        """
        Clears the subroot cache.
        """
        with self.lock:
            self.cache.clear()

        self.hits = 0
        self.misses = 0


    @abstractmethod
    def _encode_entry(self, data):
        """
        Should return the binary format of the provided data entry.

        :param data: data to encode
        :type data: whatever expected according to application logic
        :rtype: bytes
        """


    @abstractmethod
    def _store_leaf(self, data, digest):
        """
        Should create a new leaf storing the provided data entry along with
        its hash value.

        :param data: data entry
        :type data: whatever expected according to application logic
        :param digest: hashed data
        :type digest: bytes
        :returns: index of newly appended leaf counting from one
        :rtype: int
        """


    @abstractmethod
    def _get_leaf(self, index):
        """
        Should return the hash stored at the specified leaf.

        :param index: leaf index counting from one
        :type index: int
        :rtype: bytes
        """


    @abstractmethod
    def _get_leaves(self, offset, width):
        """
        Should return in respective order the hashes stored by the leaves in
        the specified range.

        :param offset: starting position counting from zero
        :type offset: int
        :param width: number of leaves to consider
        :type width: int
        :rtype: iterable of bytes
        """


    @abstractmethod
    def _get_size(self):
        """
        Should return the current number of leaves

        :rtype: int
        """


    @profile
    def _get_subroot(self, offset, width):
        """
        Cached subroot computation.

        .. warning:: The ``parameter`` is expected to be a number of two.

        :param offset: index of leftmost leaf counting from zero
        :type offset: int
        :param width: number of leaves to consider
        :type width: int
        :rtype: bytes
        """
        if width < self.threshold:
            return self._get_subroot_uncached(offset, width)

        key = (offset, width)

        with self.lock:
            try:
                value = self.cache[key]
                self.hits += 1

                return value
            except KeyError:
                pass

            self.misses += 1
            value = self._get_subroot_uncached(offset, width)
            self.cache[key] = value

        return value


    @profile
    def _get_subroot_uncached(self, offset, width):
        """
        Uncached subroot computation.

        .. warning:: The ``parameter`` is expected to be a number of two.

        :param offset: index of leftmost leaf counting from zero
        :type offset: int
        :param width: number of leaves to consider
        :type width: int
        :rtype: bytes
        """
        level = deque(self._get_leaves(offset, width))

        popleft = level.popleft
        append = level.append
        hashfunc = self.hashfunc
        prefx01 = self.prefx01
        while width > 1:
            count = 0

            while count < width:
                lnode = popleft()
                rnode = popleft()
                node = hashfunc(prefx01 + lnode + rnode).digest()
                append(node)
                count += 2

            width >>= 1

        return level[0]


    @profile
    def _get_root(self, start, limit):
        """
        Computes the root-hash for the provided leaf range.

        :param start: offset counting from zero
        :type start: int
        :param limit: last leaf index counting from one
        :type limit: int
        :rtype: bytes
        """
        subroots = deque()
        prepend = subroots.appendleft
        append = subroots.append
        pop = subroots.pop

        _get_subroot = self._get_subroot
        exponents = decompose(limit - start)
        for p in exponents:
            width = 1 << p
            offset = limit - width
            node = _get_subroot(offset, width)
            prepend(node)
            limit = offset

        hashfunc = self.hashfunc
        prefx01 = self.prefx01
        while len(subroots) > 1:
            lnode = pop()
            rnode = pop()
            node = hashfunc(prefx01 + rnode + lnode).digest()
            append(node)

        return subroots[0]


    @profile
    def _inclusion_path(self, start, offset, limit, bit):
        """
        Computes the inclusion path for the leaf located at the provided offset
        against the specified leaf range.

        .. warning:: Do not use this method directly unless you know what you
            do. Use ``prove_inclusion`` instead.

        :param start: leftmost leaf index counting from zero
        :type start: int
        :param offset: base leaf index counting from zero
        :type offset: int
        :param limit: rightmost leaf index counting from zero
        :type limit: int
        :param bit: indicates direction during path parenthetization
        :type bit: int
        :rtype: (list[int], list[bytes])
        """
        stack = deque()
        push = stack.append
        while limit > start + 1:
            k = 1 << log2(limit - start)
            if k == limit - start:
                k >>= 1

            if offset < start + k:
                push((bit, (start + k, limit)))
                limit = start + k
                bit = 0
            else:
                push((bit, (start, start + k)))
                start = start + k
                bit = 1

        _get_root = self._get_root
        _get_leaf = self._get_leaf

        rule = [bit]
        base = _get_leaf(offset + 1)
        path = [base]
        while stack:
            bit, args = stack.pop()
            rule += [bit]
            node = _get_root(*args)
            path += [node]

        return rule, path


    @profile
    def _consistency_path(self, start, offset, limit, bit):
        """
        Computes the consistency path for the state corresponding to the
        provided offset against the specified leaf range

        .. warning:: Do not use this method directly unless you know what you
            do. Use ``prove_consistency`` instead.

        :param start: leftmost leaf index counting from zero
        :type start: int
        :param offset: size corresponding to state under consideration
        :type offset: int
        :param limit: rightmost leaf index counting from zero
        :type limit: int
        :param bit: indicates direction during path parenthetization
        :type bit: int
        :rtype: (list[int], list[int], list[bytes])
        """
        stack = deque()
        push = stack.append
        while not offset == limit and not (offset == 0 and limit == 1):
            k = 1 << log2(limit)
            if k == limit:
                k >>= 1

            mask = 0
            if offset < k:
                push((bit, mask, (start + k, start + limit)))
                limit = k
                bit = 0
            else:
                mask = int(k == 1 << log2(k))
                push((bit, mask, (start, start + k)))
                start += k
                offset -= k
                limit -= k
                bit = 1

        _get_root = self._get_root
        _get_leaf = self._get_leaf

        if offset == limit:
            mask = 1
            base = _get_root(start, start + limit)
        else:
            mask = 0
            base = _get_leaf(start + offset + 1)

        rule = [bit]
        subset = [mask]
        path = [base]
        while stack:
            bit, mask, args = stack.pop()
            rule += [bit]
            subset += [mask]
            node = _get_root(*args)
            path += [node]

        return rule, subset, path


    @profile
    def _get_root_naive(self, start, limit):
        """
        Computes the root-hash for the provided leaf range.

        Essentially the same as RFC 9162, Section 2.

        .. warning:: This is an unoptimized recursive function intended for
        reference and testing. Use ``_get_root`` in production.

        :param start: offset counting from zero
        :type start: int
        :param limit: last leaf index counting from one
        :type limit: int
        :rtype: bytes
        """
        if limit == start:
            return sef.hash_empty()

        if limit == start + 1:
            return self._get_leaf(limit)

        k = 1 << log2(limit - start)
        if k == limit - start:
            k >>= 1

        lnode = self._get_root_naive(start, start + k)
        rnode = self._get_root_naive(start + k, limit)

        return self._hash_nodes(lnode, rnode)


    @profile
    def _inclusion_path_naive(self, start, offset, limit, bit):
        """
        Computes the inclusion path for the leaf located at the provided offset
        against the specified leaf range.

        Essentially the same as RFC 9162, Section 2.

        .. warning:: This is an unoptimized recursive function intended for
        reference and testing. Use ``_inclusion_path`` in production.

        :param start: leftmost leaf index counting from zero
        :type start: int
        :param offset: base leaf index counting from zero
        :type offset: int
        :param limit: rightmost leaf index counting from zero
        :type limit: int
        :param bit: indicates direction during path parenthetization
        :type bit: int
        :rtype: (list[int], list[bytes])
        """
        if offset == start and start == limit - 1:
            node = self._get_leaf(offset + 1)
            return [bit], [node]

        k = 1 << log2(limit - start)
        if k == limit - start:
            k >>= 1

        if offset < start + k:
            rule, path = self._inclusion_path_naive(start, offset, start + k, 0)
            node = self._get_root(start + k, limit)
        else:
            rule, path = self._inclusion_path_naive(start + k, offset, limit, 1)
            node = self._get_root(start, start + k)

        return rule + [bit], path + [node]


    @profile
    def _consistency_path_naive(self, start, offset, limit, bit):
        """
        Computes the consistency path for the state corresponding to the
        provided offset against the specified leaf range.

        Essentially the same as RFC 9162, Section 2.

        .. warning:: This is an unoptimized recursive function intended for
        reference and testing. Use ``_consistency_path`` in production.

        :param start: leftmost leaf index counting from zero
        :type start: int
        :param offset: size corresponding to state under consideration
        :type offset: int
        :param limit: rightmost leaf index counting from zero
        :type limit: int
        :param bit: indicates direction during path parenthetization
        :type bit: int
        :rtype: (list[int], list[int], list[bytes])
        """
        if offset == limit:
            node = self._get_root(start, start + limit)
            return [bit], [1], [node]

        if offset == 0 and limit == 1:
            node = self._get_leaf(start + offset + 1)
            return [bit], [0], [node]

        k = 1 << log2(limit)
        if k == limit:
            k >>= 1
        mask = 0

        if offset < k:
            rule, subset, path = self._consistency_path_naive(start, offset, k, 0)
            node = self._get_root(start + k, start + limit)
        else:
            rule, subset, path = self._consistency_path_naive(start + k, offset - k,
                    limit - k, 1)
            node = self._get_root(start, start + k)
            mask = int(k == 1 << log2(k))

        return rule + [bit], subset + [mask], path + [node]



import pytest
import hashlib
from pymerkle.hasher import MerkleHasher
from pymerkle.constants import ALGORITHMS
from tests.conftest import option, all_configs


data = b'oculusnonviditnecaurisaudivit'

prefx00 = b'\x00'
prefx01 = b'\x01'


@pytest.mark.parametrize('config', all_configs(option))
def test_hash_buff(config):
    h = MerkleHasher(**config)

    if h.security:
        assert h.hash_buff(data) == getattr(hashlib, h.algorithm)(
            prefx00 + data).digest()
    else:
        assert h.hash_buff(data) == getattr(hashlib, h.algorithm)(
            data).digest()


@pytest.mark.parametrize('config', all_configs(option))
def test_hash_pair(config):
    h = MerkleHasher(**config)

    if h.security:
        assert h.hash_pair(data, data) == getattr(hashlib, h.algorithm)(
            prefx01 + data + data).digest()
    else:
        assert h.hash_pair(data, data) == getattr(hashlib, h.algorithm)(
            data + data).digest()



from random import randint
import pytest

from pymerkle import SqliteTree as MerkleTree
from .conftest import option

defaults = {'warmup_rounds': 0, 'rounds': option.rounds}


opts = {'disable_optimizations': option.disable_optimizations,
        'disable_cache': option.disable_cache,
        'threshold': option.threshold,
        'capacity': option.capacity}

tree = MerkleTree(option.dbfile, algorithm=option.algorithm, **opts)


def test_root(benchmark):

    def setup():
        start = randint(0, option.size - 2) if option.randomize else 0
        final = randint(start + 1, option.size) if option.randomize \
            else option.size

        return (start, final), {}

    benchmark.pedantic(tree._get_root, setup=setup, **defaults)


def test_state(benchmark):

    def setup():
        size = randint(1, option.size) if option.randomize \
            else option.size

        return (size,), {}

    benchmark.pedantic(tree.get_state, setup=setup, **defaults)


def test_inclusion(benchmark):

    def setup():
        size = option.size
        index = randint(1, size) if option.randomize else option.index

        return (index, size), {}

    benchmark.pedantic(tree.prove_inclusion, setup=setup, **defaults)


def test_consistency(benchmark):

    def setup():
        size2 = option.size
        size1 = randint(1, size2) if option.randomize else option.index

        return (size1, size2), {}

    benchmark.pedantic(tree.prove_consistency, setup=setup, **defaults)



import pytest
from itertools import combinations
from pymerkle.utils import log2, decompose


numbers = [0]
logarithms = [0]
collections = [[]]

# 0 <= p <= 7
for logarithm in range(0, 8):

   # [0, 1, ..., p - 1]
    exponent_range = range(logarithm)

    for k in range(logarithm + 1):
        for _ in combinations(exponent_range, k):

            # [p, i_k, ..., i_1], i_1 < ... < i_k < p
            exponents = list(_) + [logarithm]

            # 2 ^ i_1 + ... + 2 ^ i_k + 2 ^ p
            number = sum(1 << i for i in exponents)

            numbers += [number]
            logarithms += [logarithm]
            collections += [exponents]


@pytest.mark.parametrize('number, logarithm', zip(numbers, logarithms))
def test_log2(number, logarithm):
    assert log2(number) == logarithm


@pytest.mark.parametrize('number, exponents', zip(numbers, collections))
def test_decompose(number, exponents):
    assert decompose(number) == exponents
    assert log2(number) == (exponents[-1] if number > 0 else 0)



from pymerkle.utils import decompose
from pymerkle.core import BaseMerkleTree


class Node:
    """
    Merkle-tree node.

    :param digest: hash value to be stored
    :type digest: bytes
    :param left: [optional] left child
    :type left: Node
    :param right: [optional] right child
    :type right: Node
    :rtype: Node
    """

    __slots__ = ('digest', 'left', 'right', 'parent')


    def __init__(self, digest, left=None, right=None):
        self.digest = digest

        self.left = left
        if left:
            left.parent = self

        self.right = right
        if right:
            right.parent = self

        self.parent = None


    def is_root(self):
        """
        Returns *True* iff the node is currently root.

        :rtype: bool
        """
        return not self.parent


    def is_leaf(self):
        """
        Returns *True* iff the node is leaf.

        :rtype: bool
        """
        return not self.left and not self.right


    def is_left_child(self):
        """
        Returns *True* iff the node is currently left child.

        :rtype: bool
        """
        parent = self.parent
        if not parent:
            return False

        return self == parent.left


    def is_right_child(self):
        """
        Returns *True* iff the node is currently right child.

        :rtype: bool
        """
        parent = self.parent
        if not parent:
            return False

        return self == parent.right


    def get_ancestor(self, degree):
        """
        .. note:: Ancestor of degree 0 is the node itself, ancestor of degree
            1 is the node's parent, etc.

        :type degree: int
        :rtype: Node
        """
        curr = self
        while degree > 0:
            curr = curr.parent
            degree -= 1

        return curr


    def expand(self, indent=2, trim=None, level=0, ignored=None):
        """
        Returns a string representing the subtree rooted at the present node.

        :param indent: [optional]
        :type indent: str
        :param trim: [optional]
        :type trim: str
        :param level: [optional]
        :type level: str
        :param ignored: [optional]
        :type ignored: str
        :rtype: str
        """
        ignored = ignored or []

        if level == 0:
            out = 2 * '\n' + ' └─' if not self.parent else ''
        else:
            out = (indent + 1) * ' '

        col = 1
        while col < level:
            out += ' │' if col not in ignored else 2 * ' '
            out += indent * ' '
            col += 1

        if self.is_left_child():
            out += ' ├──'

        if self.is_right_child():
            out += ' └──'
            ignored += [level]

        checksum = self.digest.hex()
        out += (checksum[:trim] + '...') if trim else checksum
        out += '\n'

        if self.is_leaf():
            return out

        recursion = (indent, trim, level + 1, ignored[:])

        out += self.left.expand(*recursion)
        out += self.right.expand(*recursion)

        return out


class Leaf(Node):
    """
    Merkle-tree leaf node.

    :param data: data stored by the leaf
    :type data: bytes
    :param digest: hash value stored by the leaf
    :type digest: bytes
    """

    def __init__(self, data, digest):
        self.data = data

        super().__init__(digest, None, None)


class InmemoryTree(BaseMerkleTree):
    """
    Non-persistent Merkle-tree with interior nodes loaded into the runtime.

    Inserted data is expected to be in binary format and hashed without
    further processing.

    .. warning:: This is a very memory inefficient implementation. Use it
        for debugging, testing and investigating the tree structure.
    """

    def __init__(self, algorithm='sha256', **opts):
        self.root = None
        self.leaves = []

        super().__init__(algorithm, **opts)


    def __str__(self, indent=2, trim=8):
        """
        :returns: visual representation of the tree
        :rtype: str
        """
        if not self.root:
            return '\n └─[None]\n'

        return self.root.expand(indent, trim) + '\n'


    def _encode_entry(self, data):
        """
        Returns the binary format of the provided data entry.

        :param data: data to encode
        :type data: bytes
        :rtype: bytes
        """
        return data


    def _store_leaf(self, data, digest):
        """
        Creates a new leaf storing the provided data entry along with
        its hash value.

        :param data: data entry
        :type data: whatever expected according to application lontry
        :param digest: hashed data
        :type digest: bytes
        :returns: index of newly appended leaf counting from one
        :rtype: int
        """
        tail = Leaf(data, digest)

        if not self.leaves:
            self.leaves += [tail]
            self.root = tail
            return 1

        node = self._get_last_maximal_subroot()
        self.leaves += [tail]

        digest = self._hash_nodes(node.digest, tail.digest)
        if node.is_root():
            self.root = Node(digest, node, tail)
            index = self._get_size()
            return index

        curr = node.parent
        curr.right = Node(digest, node, tail)
        curr.right.parent = curr
        while curr:
            curr.digest = self._hash_nodes(
                curr.left.digest, curr.right.digest)
            curr = curr.parent

        index = self._get_size()
        return index


    def _get_leaf(self, index):
        """
        Returns the hash stored at the specified leaf.

        :param index: leaf index counting from one
        :type index: int
        :rtype: bytes
        """
        if index < 1 or index > len(self.leaves):
            raise ValueError("%d not in leaf range" % index)

        return self.leaves[index - 1].digest


    def _get_leaves(self, offset, width):
        """
        Returns in respective order the hashes stored by the leaves in the
        specified range.

        :param offset: starting position counting from zero
        :type offset: int
        :param width: number of leaves to consider
        :type width: int
        """
        return [l.digest for l in self.leaves[offset: offset + width]]


    def _get_size(self):
        """
        :returns: current number of leaves
        :rtype: int
        """
        return len(self.leaves)


    @classmethod
    def init_from_entries(cls, entries, algorithm='sha256', **opts):
        """
        Create tree from initial data

        :param entries: initial data to append
        :type entries: iterable of bytes
        :param algorithm: [optional] hash function. Defaults to *sha256*
        :type algorithm: str
        """
        tree = cls(algorithm, **opts)

        append_entry = tree.append_entry
        for data in entries:
            append_entry(data)

        return tree


    def get_state(self, size=None):
        """
        Computes the root-hash of the subtree corresponding to the provided
        size

        .. note:: Overrides the function inherited from the base class.

        :param size: [optional] number of leaves to consider. Defaults to
            current tree size.
        :type size: int
        :rtype: bytes
        """
        currsize = self._get_size()

        if size is None:
            size = currsize

        if size == 0:
            return self.hash_empty()

        if size == currsize:
            return self.root.digest

        subroots = self._get_subroots(size)
        result = subroots[0].digest
        i = 0
        while i < len(subroots) - 1:
            result = self._hash_nodes(subroots[i + 1].digest, result)
            i += 1

        return result


    def _inclusion_path_fallback(self, offset):
        """
        Non-recursive utility using concrete traversals to compute the inclusion
        path against the current number of leaves.

        :param offset: base leaf index counting from zero
        :type offset: int
        :rtype: (list[int], list[bytes])
        """
        base = self.leaves[offset]
        bit = 1 if base.is_right_child() else 0

        path = [base.digest]
        rule = [bit]

        curr = base
        while curr.parent:
            parent = curr.parent

            if curr.is_left_child():
                digest = parent.right.digest
                bit = 0 if parent.is_left_child() else 1
            else:
                digest = parent.left.digest
                bit = 1 if parent.is_right_child() else 0

            rule += [bit]
            path += [digest]
            curr = parent

        # Last bit is insignificant; fix it to zero just to be fully compatible
        # with the output of the overriden method
        rule[-1] = 0

        return rule, path


    def _inclusion_path(self, start, offset, limit, bit):
        """
        Computes the inclusion path for the leaf located at the provided offset
        against the specified leaf range

        .. warning:: This is an unoptimized recursive function intended for
        reference and testing. Use ``_inclusion_path`` in production.

        :param start: leftmost leaf index counting from zero
        :type start: int
        :param offset: base leaf index counting from zero
        :type offset: int
        :param limit: rightmost leaf index counting from zero
        :type limit: int
        :param bit: indicates direction during path parenthetization
        :type bit: int
        :rtype: (list[int], list[bytes])
        """
        if start == 0 and limit == self._get_size():
            return self._inclusion_path_fallback(offset)

        return super()._inclusion_path(start, offset, limit, bit)


    def _get_subroot_node(self, index, height):
        """
        Returns the root node of the perfect subtree of the provided height whose
        leftmost leaf node is located at the provided position.

        .. note:: Returns *None* if no binary subtree exists for the provided
            parameters.

        :param index: position of leftmost leaf node coutning from one
        :type index: int
        :param height: height of requested subtree
        :type height: int
        :rtype: Node
        """
        node = self.leaves[index - 1]

        if not node:
            return

        i = 0
        while i < height:
            curr = node.parent

            if not curr:
                return

            if curr.left is not node:
                return

            node = curr
            i += 1

        # Verify existence of perfect subtree rooted at the detected node
        curr = node
        i = 0
        while i < height:
            if curr.is_leaf():
                return

            curr = curr.right
            i += 1

        return node


    def _get_last_maximal_subroot(self):
        """
        Returns the root node of the perfect subtree of maximum possible size
        containing the currently last leaf.

        :rtype: Node
        """
        degree = decompose(len(self.leaves))[0]

        return self.leaves[-1].get_ancestor(degree)


    def _get_subroots(self, size):
        """
        Returns in respective order the root nodes of the successive perfect
        subtrees whose sizes sum up to the provided size.

        :param size:
        :type size: int
        :rtype: list[Node]
        """
        if size < 0 or size > self._get_size():
            return []

        subroots = []
        offset = 0
        for height in reversed(decompose(size)):
            node = self._get_subroot_node(offset + 1, height)

            if not node:
                return []

            subroots += [node]
            offset += 1 << height

        return list(reversed(subroots))



import hashlib
from pymerkle import constants


class MerkleHasher:
    """
    Encapsulates elementary hashing operations.

    :param algorithm: hash algorithm
    :type algorithm: str
    :param security: [optional] resistance against second-preimage attack. Defaults
        to *True*
    :type security: bool
    """

    def __init__(self, algorithm, security=True, **kw):
        normalized = algorithm.lower().replace('-', '_')

        if normalized not in constants.ALGORITHMS:
            raise ValueError(f'{algorithm} not supported')

        self.prefx00 = b'\x00' if security else b''
        self.prefx01 = b'\x01' if security else b''

        self.algorithm = algorithm
        self.security = security
        self.hashfunc = getattr(hashlib, algorithm)


    def _consume_bytes(self, buff):
        """
        :param buff:
        :type buff: bytes
        :rtype: bytes
        """
        hasher = self.hashfunc()
        update = hasher.update
        chunksize = 1024
        offset = 0
        chunk = buff[offset: chunksize]
        while chunk:
            update(chunk)
            offset += chunksize
            chunk = buff[offset: offset + chunksize]

        return hasher.digest()


    def hash_empty(self):
        """
        Computes the hash of the empty data without prepending security
        prefixes.

        :param buff:
        :type buff: bytes
        :rtype: bytes
        """
        return self._consume_bytes(b'')


    def hash_raw(self, buff):
        """
        Computes the hash of the provided data without prepending security
        prefixes.

        :param buff:
        :type buff: bytes
        :rtype: bytes
        """
        return self._consume_bytes(buff)


    def hash_buff(self, data):
        """
        Computes the hash of the provided binary data.

        .. note:: Prepends ``\\x00`` if security mode is enabled

        :type data: bytes
        :rtype: bytes
        """
        return self._consume_bytes(self.prefx00 + data)



    def hash_pair(self, buff1, buff2):
        """
        Computes the hash of the concatenation of the provided binary data.

        .. note:: Prepends ``\\x01`` if security mode is enabled

        :param buff1: left value
        :type buff1: bytes
        :param buff2: right value
        :type buff2: bytes
        :rtype: bytes
        """
        return self.hashfunc(self.prefx01 + buff1 + buff2).digest()



import os
import math
import pytest

from pymerkle import SqliteTree as MerkleTree

current_dir = os.path.dirname(os.path.abspath(__file__))

DEFAULT_DB = os.path.join(current_dir, 'merkle.db')
DEFAULT_SIZE = 10 ** 6
DEFAULT_INDEX = math.ceil(DEFAULT_SIZE / 2)
DEFAULT_ROUNDS = 100
DEFAULT_THRESHOLD = 128
DEFAULT_CAPACITY = 1024 ** 3


def pytest_addoption(parser):
    parser.addoption('--dbfile', type=str, default=DEFAULT_DB,
        help='Database filepath')
    parser.addoption('--size', type=int, default=DEFAULT_SIZE,
        help='Nr entries to consider')
    parser.addoption('--index', type=int, default=DEFAULT_INDEX,
        help='Base index for proof operations')
    parser.addoption('--rounds', type=int, default=DEFAULT_ROUNDS,
        help='Nr rounds per benchmark')
    parser.addoption('--algorithm', default='sha256',
        help='Hash algorithm used by the tree')
    parser.addoption('--randomize', action='store_true', default=False,
        help='Randomize function input per round')
    parser.addoption('--disable-optimizations', action='store_true', default=False,
        help='Use unoptimized versions of core operations')
    parser.addoption('--disable-cache', action='store_true', default=False,
        help='Disable subroot caching')
    parser.addoption('--threshold', type=int, metavar='WIDTH',
        default=DEFAULT_THRESHOLD,
        help='Subroot cache threshold')
    parser.addoption('--capacity', type=int, metavar='BYTES',
        default=DEFAULT_CAPACITY,
        help='Subroot cache capacity in bytes')

option = None

def pytest_configure(config):
    global option
    option = config.option



from itertools import product
import pytest

from pymerkle.utils import decompose
from tests.conftest import option, tree_and_index, tree_and_range


@pytest.mark.parametrize('tree, start, limit', tree_and_range())
def test_get_root(tree, start, limit):
    offset = start
    subroots = []
    for p in reversed(decompose(limit - start)):
        node = tree._get_subroot(offset, 1 << p)
        subroots += [node]
        offset += 1 << p

    subroots = list(reversed(subroots))
    root = subroots[0]
    index = 0
    while index < len(subroots) - 1:
        root = tree._hash_nodes(subroots[index + 1], root)
        index += 1

    assert root == tree._get_root_naive(start, limit)
    assert root == tree._get_root(start, limit)


@pytest.mark.parametrize('tree, size', tree_and_index())
def test_state(tree, size):
    state = tree.get_state(size)

    assert state == tree._get_root_naive(0, size)
    assert state == tree._get_root(0, size)


@pytest.mark.parametrize('tree, start, limit', tree_and_range())
def test_inclusion_path(tree, start, limit):
    for bit, offset in product([0, 1], range(start, limit)):
        path1 = tree._inclusion_path(start, offset, limit, bit)
        path2 = tree._inclusion_path_naive(start, offset, limit, bit)

        assert path1 == path2


@pytest.mark.parametrize('tree, size1, size2', tree_and_range())
def test_consistency_path(tree, size1, size2):
    for bit1, bit2 in product([0, 1], [0, 1]):
        path1 = tree._consistency_path(bit1, size1, size2, bit2)
        path2 = tree._consistency_path_naive(bit1, size1, size2, bit2)

        assert path1 == path2



import sqlite3
from pymerkle.core import BaseMerkleTree


class SqliteTree(BaseMerkleTree):
    """
    Persistent Merkle-tree implementation using a SQLite database as storage.

    Inserted data is expected to be in binary format and hashed without
    further processing.

    .. note:: The database schema consists of a single table called *leaf*
        with two columns: *index*, which is the primary key serving as leaf
        index, and *entry*, which is a blob field storing the appended data.

    :param dbfile: database filepath
    :type dbfile: str
    :param algorithm: [optional] hashing algorithm. Defaults to *sha256*
    :type algorithm: str
    """

    def __init__(self, dbfile, algorithm='sha256', **opts):
        self.dbfile = dbfile
        self.con = sqlite3.connect(self.dbfile)
        self.con.row_factory = lambda cursor, row: row[0]
        self.cur = self.con.cursor()

        with self.con:
            query = f'''
                CREATE TABLE IF NOT EXISTS leaf(
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    entry BLOB,
                    hash BLOB
                );'''
            self.cur.execute(query)

        super().__init__(algorithm, **opts)


    def __enter__(self):
        return self


    def __exit__(self, *exc):
        self.con.close()


    def _encode_entry(self, data):
        """
        Returns the binary format of the provided data entry.

        :param data: data to encode
        :type data: bytes
        :rtype: bytes
        """
        return data


    def _store_leaf(self, data, digest):
        """
        Creates a new leaf storing the provided data along with its
        hash value.

        :param data: data entry
        :type data: whatever expected according to application logic
        :param digest: hashed data
        :type digest: bytes
        :returns: index of newly appended leaf counting from one
        :rtype: int
        """
        if not isinstance(data, bytes):
            raise ValueError('Provided data is not binary')

        cur = self.cur

        with self.con:
            query = f'''
                INSERT INTO leaf(entry, hash) VALUES (?, ?)
            '''
            cur.execute(query, (data, digest))

        return cur.lastrowid


    def _get_leaf(self, index):
        """
        Returns the hash stored at the specified leaf.

        :param index: leaf index counting from one
        :type index: int
        :rtype: bytes
        """
        cur = self.cur

        query = f'''
            SELECT hash FROM leaf WHERE id = ?
        '''
        cur.execute(query, (index,))

        return cur.fetchone()


    def _get_leaves(self, offset, width):
        """
        Returns in respective order the hashes stored by the leaves in the
        specified range.

        :param offset: starting position counting from zero
        :type offset: int
        :param width: number of leaves to consider
        :type width: int
        """
        cur = self.cur

        query = f'''
            SELECT hash FROM leaf WHERE id BETWEEN ? AND ?
        '''
        cur.execute(query, (offset + 1, offset + width))

        return cur.fetchall()


    def _get_size(self):
        """
        :returns: current number of leaves
        :rtype: int
        """
        cur = self.cur

        query = f'''
            SELECT COUNT(*) FROM leaf
        '''
        cur.execute(query)

        return cur.fetchone()


    def get_entry(self, index):
        """
        Returns the unhashed data stored at the specified leaf.

        :param index: leaf index counting from one
        :type index: int
        :rtype: bytes
        """
        cur = self.cur

        query = f'''
            SELECT entry FROM leaf WHERE id = ?
        '''
        cur.execute(query, (index,))

        return cur.fetchone()


    def _hash_per_chunk(self, entries, chunksize):
        """
        Generator yielding in chunks pairs of entry data and hash value.

        :param entries:
        :type entries: iterable of bytes
        :param chunksize:
        :type chunksize: int
        """
        _hash_entry = self.hash_buff

        offset = 0
        chunk = entries[offset: chunksize]
        while chunk:
            hashes = [_hash_entry(data) for data in chunk]
            yield zip(chunk, hashes)

            offset += chunksize
            chunk = entries[offset: offset + chunksize]


    def append_entries(self, entries, chunksize=100_000):
        """
        Bulk operation for appending a batch of entries.

        :param entries: data entries to append
        :type entries: iterable of bytes
        :param chunksize: [optional] number entries to insert per
            database transaction.
        :type chunksize: int
        :returns: index of last appended entry
        :rtype: int
        """
        cur = self.cur

        with self.con:
            query = f'''
                INSERT INTO leaf(entry, hash) VALUES (?, ?)
            '''
            for chunk in self._hash_per_chunk(entries, chunksize):
                cur.execute('BEGIN TRANSACTION')

                for (data, digest) in chunk:
                    cur.execute(query, (data, digest))

                cur.execute('END TRANSACTION')

        return cur.lastrowid



"""
Run tree operations for profiling purposes
"""

import os
import sys
import argparse
from random import randint

from pymerkle import SqliteTree as MerkleTree
from pymerkle import constants

current_dir = os.path.dirname(os.path.abspath(__file__))
parent_dir = os.path.dirname(current_dir)

DEFAULT_DB = os.path.join(parent_dir, 'benchmarks', 'merkle.db')
DB_SIZE = 10 ** 6
DEFAULT_ROUNDS = 1
DEFAULT_THRESHOLD = 128
DEFAULT_CAPACITY = 1024 ** 3


def parse_cli_args():
    config = {'prog': sys.argv[0], 'usage': 'python %s' % sys.argv[0],
              'description': __doc__, 'epilog': '\n',
              'formatter_class': argparse.ArgumentDefaultsHelpFormatter}
    parser = argparse.ArgumentParser(**config)

    parser.add_argument('--dbfile', type=str, default=DEFAULT_DB,
        help='Database filepath')
    parser.add_argument('--algorithm', choices=constants.ALGORITHMS,
        default='sha256', help='Hashing algorithm')
    parser.add_argument('--rounds', type=int, default=DEFAULT_ROUNDS,
        help='Nr rounds')
    parser.add_argument('--randomize', action='store_true', default=False,
        help='Randomize function input per round')
    parser.add_argument('--disable-optimizations', action='store_true', default=False,
        help='Use unoptimized versions of core functionalities')
    parser.add_argument('--disable-cache', action='store_true', default=False,
        help='Disable subroot caching')
    parser.add_argument('--threshold', type=int, metavar='WIDTH',
        default=DEFAULT_THRESHOLD, 
        help='Subroot cache threshold')
    parser.add_argument('--capacity', type=int, metavar='BYTES',
        default=DEFAULT_CAPACITY,
        help='Subroot cache capacity in bytes')

    operation = parser.add_subparsers(dest='operation')

    root = operation.add_parser('root',
        help='Run `_get_root`')
    root.add_argument('--start', type=int, default=0,
        help='Starting position')
    root.add_argument('--limit', type=int, default=DB_SIZE,
        help='Final position')

    state = operation.add_parser('state',
        help='Run `get_state`')
    state.add_argument('--size', type=int, default=DB_SIZE,
        help='Nr entries to consider')

    inclusion = operation.add_parser('inclusion',
        help='Run `prove_inclusion`')
    inclusion.add_argument('--index', type=int, required=True,
        help='Leaf index')
    inclusion.add_argument('--size', type=int, default=DB_SIZE,
        help='Nr entries to consider')

    consistency = operation.add_parser('consistency',
        help='Run `prove_consistency`')
    consistency.add_argument('--size1', type=int, required=True,
        help='Size of prior state')
    consistency.add_argument('--size2', type=int, default=DB_SIZE,
        help='Size of later state')

    return parser.parse_args()


if __name__ == '__main__':
    cli = parse_cli_args()

    opts = {'disable_optimizations': cli.disable_optimizations,
            'disable_cache': cli.disable_cache,
            'threshold': cli.threshold,
            'capacity': cli.capacity}

    tree = MerkleTree(cli.dbfile, algorithm=cli.algorithm, **opts)

    match cli.operation:
        case 'root':
            func = tree._get_root

            def get_args():
                return (cli.start, cli.limit)

            if cli.randomize:
                def get_args():
                    start = randint(0, cli.limit - 2)
                    limit = randint(start + 1, cli.limit)

                    return (start, limit)

        case 'state':
            func = tree.get_state

            def get_args():
                return (cli.size,)

            if cli.randomize:
                def get_args():
                    size = randint(1, cli.size)

                    return (size,)

        case 'inclusion':
            func = tree.prove_inclusion

            def get_args():
                return (cli.index, cli.size)

            if cli.randomize:
                def get_args():
                    size = cli.size
                    index = randint(1, size)

                    return (index, size)

        case 'consistency':
            func = tree.prove_consistency

            def get_args():
                return (cli.size1, cli.size2)

            if cli.randomize:
                def get_args():
                    size2 = cli.size2
                    size1 = randint(1, size2)

                    return (size1, size2)

    count = 0
    while count < cli.rounds:
        args = get_args()
        print('round %d:' % count, args)
        func(*args)
        count += 1

    print("\033[92m {}\033[00m".format(tree.get_cache_info()))



"""
Performs second-preimage attack for both security modes. Attach should succeed
only when the tree's securitu mode has been disabled, that is, iff the security
attribute has been set to false

Attack Schema
-------------

               Original tree               Attacker's tree

                     G                           G
                   /   \                       /   \
                 E       F = h(CD)           E      F    ------> [injected leaf]
                / \     / \                 / \     |
hashes:        A   B   C   D               A   B   (CD)  <------ [forged entry]
               |   |   |   |               |   |
entries:       a   b   c   d               a   b


Concatenate the 3-rd and 4-th leaf-hashes and append the result as 3-rd,
leaving the rest leaves unoutched
"""

import pytest
from tests.conftest import all_configs, option, resolve_backend

MerkleTree = resolve_backend(option)


@pytest.mark.parametrize('config', all_configs(option))
def test_second_preimage_attack(config):
    tree = MerkleTree.init_from_entries([b'foo', b'bar', b'baz', b'qux'],
        **config)

    forged = tree.get_leaf(3) + tree.get_leaf(4)
    attacker = MerkleTree.init_from_entries([b'foo', b'bar', forged],
        **config)

    assert tree.security ^ (attacker.get_state() == tree.get_state())



"""
Create tree sqlite database for benchmarking. WARNING: Unless otherwise specified,
the database file will be overwritten if it already exists.
"""

import os
import sys
import argparse
import time

from pymerkle import SqliteTree, constants

current_dir = os.path.dirname(os.path.abspath(__file__))

DEFAULT_DB = os.path.join(current_dir, 'merkle.db')
DEFAULT_ALGORITHM = 'sha256'
DEFAULT_SIZE = 10 ** 8
DEFAULT_BATCHSIZE = 10 ** 7


def parse_cli_args():
    config = {'prog': sys.argv[0], 'usage': 'python %s' % sys.argv[0],
              'description': __doc__, 'epilog': '\n',
              'formatter_class': argparse.ArgumentDefaultsHelpFormatter}
    parser = argparse.ArgumentParser(**config)

    parser.add_argument('--dbfile', type=str, default=DEFAULT_DB,
        help='Database filepath')
    parser.add_argument('--algorithm', choices=constants.ALGORITHMS,
        default=DEFAULT_ALGORITHM, help='Hashing algorithm')
    parser.add_argument('--disable-security', action='store_true', default=False,
        help='Disable resistance against 2nd-preimage attack')
    parser.add_argument('--size', type=int, default=DEFAULT_SIZE,
        help='Nr entries to append in total')
    parser.add_argument('--batchsize', type=int, default=DEFAULT_BATCHSIZE,
        help='Nr entries to append per bulk insertion')
    parser.add_argument('--preserve-database', action='store_true', default=False,
        help='Append without overwriting if already existent')

    return parser.parse_args()


if __name__ == '__main__':
    args = parse_cli_args()

    batchsize = args.batchsize
    size = args.size
    if batchsize > size:
        sys.stdout.write("[-] Batchsize exceeds size\n")
        sys.exit(1)

    if not args.preserve_database:
        try:
            os.remove(args.dbfile)
        except OSError:
            pass

    opts = {'algorithm': args.algorithm,
            'disable_security': args.disable_security}

    with SqliteTree(args.dbfile, **opts) as tree:
        offset = 0
        count = 1
        append_entries = tree.append_entries
        chunksize = min(100_000, batchsize)
        start_time = time.time()
        while offset < size:
            limit = offset + batchsize + 1
            if limit > size + 1:
                limit = size + 1

            print(f"\nCreating {batchsize} entries...")
            entries = [f'entry-{i}'.encode('utf-8') for i in range(offset + 1,
                limit)]

            index = append_entries(entries, chunksize)
            assert index == limit - 1

            currsize = tree.get_size()
            print("\033[92m {}\033[00m".format(f"Appended batch {count}"))
            print("\033[92m {}\033[00m".format(f"Current size: {currsize}"))

            count += 1
            offset += batchsize

        end_time = time.time()
        elapsed_time = end_time - start_time

        assert currsize == args.size

        print(f"\nDatabase at {args.dbfile}\n")
        print("\nTime elapsed (sec):", elapsed_time)



import pytest
from tests.conftest import tree_and_index

from pymerkle import verify_inclusion, verify_consistency, InvalidChallenge, \
    InvalidProof


@pytest.mark.parametrize('tree, size1', tree_and_index())
def test_consistency_success(tree, size1):
    size2 = tree.get_size()
    proof = tree.prove_consistency(size1, size2)

    state1 = tree.get_state(size1)
    state2 = tree.get_state()

    verify_consistency(state1, state2, proof)


@pytest.mark.parametrize('tree, size1', tree_and_index(default_config=True))
def test_consistency_invalid_prior(tree, size1):
    size2 = tree.get_size()
    proof = tree.prove_consistency(size1, size2)

    state1 = tree.hash_raw(b'random')
    state2 = tree.get_state(size2)
    with pytest.raises(InvalidProof):
        verify_consistency(state1, state2, proof)


@pytest.mark.parametrize('tree, size1', tree_and_index(default_config=True))
def test_consistency_invalid_state(tree, size1):
    size2 = tree.get_size()
    proof = tree.prove_consistency(size1, size2)

    state1 = tree.get_state(size1)
    state2 = tree.hash_raw(b'random')
    with pytest.raises(InvalidProof):
        verify_consistency(state1, state2, proof)


@pytest.mark.parametrize('tree, size1', tree_and_index(default_config=True))
def test_consistency_invalid_challenge(tree, size1):
    size2 = tree.get_size()

    with pytest.raises(InvalidChallenge):
        tree.prove_consistency(size1, size2 + 1)

    with pytest.raises(InvalidChallenge):
        tree.prove_consistency(size1, -1)

    with pytest.raises(InvalidChallenge):
        tree.prove_consistency(size1 + 1, size1)

    with pytest.raises(InvalidChallenge):
        tree.prove_consistency(-1, size1)



import os
import json
from time import time

from pymerkle.hasher import MerkleHasher


class InvalidProof(Exception):
    """
    Raised when a Merkle-proof is found to be invalid.
    """
    pass


def verify_inclusion(base, root, proof):
    """
    Verifies the provided Merkle-proof of inclusion against the provided leaf
    hash and tree state.

    :param base: acclaimed leaf hash
    :type base: bytes
    :param root: acclaimed root hash
    :type root: bytes
    :param proof: proof of inclusion
    :type proof: MerkleProof
    :raises InvalidProof: if the proof is found invalid
    """
    if not proof.path[0] == base:
        raise InvalidProof('Base hash does not match')

    if not proof.resolve() == root:
        raise InvalidProof('State does not match')


def verify_consistency(state1, state2, proof):
    """
    Verifies the provided Merkle-proof of consistency against the given states.

    :param state1: acclaimed prior state
    :type state1: bytes
    :param state2: acclaimed later state
    :type state2: bytes
    :raises InvalidProof: if the proof is found invalid
    :param proof: proof of consistency
    :type proof: MerkleProof
    """
    if not proof.retrieve_prior_state() == state1:
        raise InvalidProof('Prior state does not match')

    if not proof.resolve() == state2:
        raise InvalidProof('Later state does not match')


class MerkleProof:
    """
    Verifiable Merkle-proof object

    :param algorithm: hash algorithm to be applied during state resolution
    :type algorithm: str
    :param security: resistance against 2-nd preimage attack indicator
    :type security: bool
    :param size: tree size corresponding to requested state
    :type size: int
    :param rule: specifies parenthetization of hashes during state
        resolution
    :type rule: list[int]
    :param subset: indicates subset of hashes during prior state resolution
        (makes sense only for consistency proofs)
    :type subset: list[int]
    :param path: path of hashes
    :type path: list[bytes]
    """

    def __init__(self, algorithm, security, size, rule, subset, path):
        self.algorithm = algorithm
        self.security = security
        self.size = size
        self.rule = rule
        self.subset = subset
        self.path = path
        self.hasher = MerkleHasher(**self.get_metadata())


    def get_metadata(self):
        """
        Returns the information needed to configure the hashing machinery.

        :rtype: dict
        """
        return {'algorithm': self.algorithm, 'security': self.security,
                'size': self.size}

    def serialize(self):
        """
        Returns the JSON representation of the verifiable object.

        :rtype: dict
        """
        return {
            'metadata': {
                'algorithm': self.algorithm,
                'security': self.security,
                'size': self.size,
            },
            'rule': self.rule,
            'subset': self.subset,
            'path': [digest.hex() for digest in self.path]
        }


    @classmethod
    def deserialize(cls, data):
        """
        :param data:
        :type data: dict
        :rtype: MerkleProof
        """
        metadata = data['metadata']
        rule = data['rule']
        subset = data['subset']
        path = [bytes.fromhex(checksum) for checksum in data['path']]

        return cls(**metadata, rule=rule, subset=subset, path=path)


    def retrieve_prior_state(self):
        """
        Computes the acclaimed prior state as specified by the included path of
        hashes.

        .. note:: Makes sense only for consistency proofs.

        :rtype: bytes
        """
        subpath = [digest for (mask, digest) in zip(self.subset, self.path) if
            mask]

        if not subpath:
            return self.hasher.hash_empty()

        result = subpath[0]
        index = 0
        hash_pair = self.hasher.hash_pair
        while index < len(subpath) - 1:
            result = hash_pair(subpath[index + 1], result)
            index += 1

        return result


    def resolve(self):
        """
        Computes the target hash of the included path of hashes.

        :rtype: bytes
        """
        path = list(zip(self.rule, self.path))

        if not path:
            return self.hasher.hash_empty()

        bit, result = path[0]
        index = 0
        hash_pair = self.hasher.hash_pair
        while index < len(path) - 1:
            next_bit, digest = path[index + 1]

            match bit:
                case 0:
                    result = hash_pair(result, digest)
                case 1:
                    result = hash_pair(digest, result)
                case _:
                    raise ValuError('Invalid bit found')

            bit = next_bit
            index += 1

        return result



import pytest
from tests.conftest import tree_and_index

from pymerkle import verify_inclusion, verify_consistency, InvalidChallenge, \
    InvalidProof


@pytest.mark.parametrize('tree, index', tree_and_index())
def test_inclusion_success(tree, index):
    base = tree.get_leaf(index)
    proof = tree.prove_inclusion(index)
    state = tree.get_state()

    verify_inclusion(base, state, proof)


@pytest.mark.parametrize('tree, index', tree_and_index(default_config=True))
def test_inclusion_invalid_base(tree, index):
    base = tree.hash_raw(b'random')
    proof = tree.prove_inclusion(index)
    state = tree.get_state()

    with pytest.raises(InvalidProof):
        verify_inclusion(base, state, proof)


@pytest.mark.parametrize('tree, index', tree_and_index(default_config=True))
def test_inclusion_invalid_state(tree, index):
    base = tree.get_leaf(index)
    proof = tree.prove_inclusion(index)
    state = tree.hash_raw(b'random')

    with pytest.raises(InvalidProof):
        verify_inclusion(base, state, proof)


@pytest.mark.parametrize('tree, index', tree_and_index(default_config=True))
def test_inclusion_invalid_challenge(tree, index):
    size = tree.get_size()

    with pytest.raises(InvalidChallenge):
        tree.prove_inclusion(index, size + 1)

    with pytest.raises(InvalidChallenge):
        tree.prove_inclusion(0, size)

    with pytest.raises(InvalidChallenge):
        tree.prove_inclusion(-1, size)

    with pytest.raises(InvalidChallenge):
        tree.prove_inclusion(index + 1, index)
